/***********************************************
 *	FILE:	Accel_BMA180.cpp
 *
 *	AUTHORS:	-Guillermo Maroto
 *
 *	Based on Aeroquad.com - Accel.h         
 *   --------+--------         --------+--------
 *          | |_______________________| | 
 *                   |         |
 *                   |_________|
 *                 ~____|___|____    
 **********************************************/
#include "Accel_BMA180.h"


void Accel_BMA180::initialize(void) 
{
	byte data;
	
	this->_initialize(0,1,2);  // AKA added for consistency
	
	accelOneG        = readFloat(ACCEL1G_ADR);
	accelZero[XAXIS] = readFloat(LEVELPITCHCAL_ADR);
	accelZero[YAXIS] = readFloat(LEVELROLLCAL_ADR);
	accelZero[ZAXIS] = readFloat(LEVELZCAL_ADR);
	smoothFactor     = readFloat(ACCSMOOTH_ADR);
	
	// Check if accel is connected
	if (readWhoI2C(accelAddress) != 0x03) // page 52 of datasheet
		Serial.println("Accelerometer not found!");
	
	// Thanks to SwiftingSpeed for updates on these settings
	// http://aeroquad.com/showthread.php?991-AeroQuad-Flight-Software-v2.0&p=11207&viewfull=1#post11207
	updateRegisterI2C(accelAddress, 0x10, 0xB6); //reset device
	delay(10);  //sleep 10 ms after reset (page 25)
	
	// In datasheet, summary register map is page 21
	// Low pass filter settings is page 27
	// Range settings is page 28
	updateRegisterI2C(accelAddress, 0x0D, 0x10); //enable writing to control registers
	sendByteI2C(accelAddress, 0x20); // register bw_tcs (bits 4-7)
	data = readByteI2C(accelAddress); // get current register value
	updateRegisterI2C(accelAddress, 0x20, data & 0x0F); // set low pass filter to 10Hz (value = 0000xxxx)
	
	// From page 27 of BMA180 Datasheet
	//  1.0g = 0.13 mg/LSB
	//  1.5g = 0.19 mg/LSB
	//  2.0g = 0.25 mg/LSB
	//  3.0g = 0.38 mg/LSB
	//  4.0g = 0.50 mg/LSB
	//  8.0g = 0.99 mg/LSB
	// 16.0g = 1.98 mg/LSB
	sendByteI2C(accelAddress, 0x35); // register offset_lsb1 (bits 1-3)
	data = readByteI2C(accelAddress);
	data &= 0xF1;
	data |= 0x04; // Set range select bits for +/-2g
	updateRegisterI2C(accelAddress, 0x35, data);
}


void Accel_BMA180::measure(void) 
{
	//int rawData[3];
	
	Wire.beginTransmission(accelAddress);
	Wire.send(0x02);
	Wire.endTransmission();
	Wire.requestFrom(accelAddress, 6);
	
	//Modified by MarotoCopter
	//Curiosidad: Lo dejo asi aunque realmente el XAXIS es el eje y y viceversa, pero en flightAngle_DCM (la unica parte donde se utiliza Accel)
	//lo han tenido en cuenta de la forma erronea. Ahora mismo Roll giraria sobre el eje X y no sobre Y que es como realmente es.
	
		accelADC[YAXIS] = ((Wire.receive()|(Wire.receive() << 8)) >> 2) - accelZero[YAXIS];
		accelData[YAXIS] = filterSmooth(accelADC[YAXIS] * accelScaleFactor, accelData[YAXIS], smoothFactor);
	
		accelADC[XAXIS] = accelZero[XAXIS] - ((Wire.receive()|(Wire.receive() << 8)) >> 2);
		accelData[XAXIS] = filterSmooth(accelADC[XAXIS] * accelScaleFactor, accelData[XAXIS], smoothFactor);
	
		accelADC[ZAXIS] = accelZero[ZAXIS] - ((Wire.receive()|(Wire.receive() << 8)) >> 2);
		accelData[ZAXIS] = filterSmooth(accelADC[ZAXIS] * accelScaleFactor, accelData[ZAXIS], smoothFactor);
	
	/*
	for (byte axis = XAXIS; axis < LASTAXIS; axis++) {
		if (axis == XAXIS)
			accelADC[axis] = ((Wire.receive()|(Wire.receive() << 8)) >> 2) - accelZero[axis];
		else
			accelADC[axis] = accelZero[axis] - ((Wire.receive()|(Wire.receive() << 8)) >> 2);
		accelData[axis] = filterSmooth(accelADC[axis] * accelScaleFactor, accelData[axis], smoothFactor);
	}
	 */
}

// Allows user to zero accelerometers on command
void Accel_BMA180::calibrate(void) 
{  
    int findZero[FINDZERO];
    int dataAddress;
    
    for (byte calAxis = XAXIS; calAxis < ZAXIS; calAxis++) {
		if (calAxis == XAXIS) dataAddress = 0x02;		//Modified by MarotoCopter
		if (calAxis == YAXIS) dataAddress = 0x04;		//Modified by MarotoCopter
		if (calAxis == ZAXIS) dataAddress = 0x06;
		for (int i=0; i<FINDZERO; i++) {
			sendByteI2C(accelAddress, dataAddress);
			findZero[i] = readReverseWordI2C(accelAddress) >> 2; // last two bits are not part of measurement
			delay(10);
		}
		accelZero[calAxis] = findMedian(findZero, FINDZERO);
    }
	
    // replace with estimated Z axis 0g value
    accelZero[ZAXIS] = (accelZero[XAXIS] + accelZero[PITCH]) / 2;
    // store accel value that represents 1g
    measure();
    accelOneG = -accelData[ZAXIS];
	
    writeFloat(accelOneG,        ACCEL1G_ADR);
    writeFloat(accelZero[XAXIS], LEVELPITCHCAL_ADR);
    writeFloat(accelZero[YAXIS], LEVELROLLCAL_ADR);
    writeFloat(accelZero[ZAXIS], LEVELZCAL_ADR);
	
	Serial.println("Accel_BMA180 calibrated!");
}


