/***********************************************
 *	FILE:	Quadcpter.cpp
 *
 *	AUTHORS:	-Guillermo Maroto
 *
 *           
 *   --------+--------         --------+--------
 *          | |_______________________| | 
 *                   |         |
 *                   |_________|
 *                 ~____|___|____    
 **********************************************/
#include "Quadcopter.h"

void Quadcopter::initialize()
{
	//initializes basic things common to all
	
	//initialize internal communication .  I2C
	Wire.begin();
	// Recommendation from Mgros to increase I2C speed to 400kHz
	// http://aeroquad.com/showthread.php?991-AeroQuad-Flight-Software-v2.0&p=11262&viewfull=1#post11262
	TWBR = 12;
	
	
		control[ROLL].setP(PROLL);
		control[ROLL].setI(IROLL);
		control[ROLL].setD(DROLL);
		control[PITCH].setP(PPITCH);
		control[PITCH].setI(IPITCH);
		control[PITCH].setD(DPITCH);
		control[YAW].setP(PYAW);
		control[YAW].setI(IYAW);
		control[YAW].setD (DYAW);
		controlHeading.setP(PHEADING);
		controlHeading.setI(IHEADING);
		controlHeading.setD(DHEADING);
		rollAttitudeStableMode.setP(PROLLATTITUDE);
		rollAttitudeStableMode.setI(IROLLATTITUDE);
		pitchAttitudeStableMode.setP(PPITCHATTITUDE);
		pitchAttitudeStableMode.setI(IPITCHATTITUDE);
		
	
		gyro.setSmoothFactor(GYRO_SMOOTH_FACTOR);
		accel.setSmoothFactor(ACCEL_SMOOTH_FACTOR);
		
}

void Quadcopter::processFlightControl()
{		
	
		if(Power < 1100)
		{
			zeroIntegralError();
		}
		
	pitchAttitudeStableMode.setTargetPosition(TargetPositionUser[PITCH]);
	rollAttitudeStableMode.setTargetPosition(TargetPositionUser[ROLL]);
	
	rollAttitudeStableMode.updatePID( flightAngle.getData(ROLL));
	pitchAttitudeStableMode.updatePID(-flightAngle.getData(PITCH));
		
	control[ROLL].setTargetPosition(rollAttitudeStableMode.getOutput());
	control[PITCH].setTargetPosition(pitchAttitudeStableMode.getOutput());
		
	control[ROLL].updatePID(gyro.getData(ROLL));
	control[PITCH].updatePID(-gyro.getData(PITCH));

	control[YAW].setTargetPosition(constrain(TargetPositionUser[YAW] + processHeading(), -PI, PI));
	control[YAW].updatePID(gyro.getData(YAW));
	
	
	motor[FRONT].write(Power-control[PITCH].getOutput()-control[YAW].getOutput()+TRIMOTORFRONT);
	motor[REAR].write(Power+control[PITCH].getOutput()-control[YAW].getOutput()+TRIMOTORREAR);
	
	motor[LEFT].write(Power+control[ROLL].getOutput()+control[YAW].getOutput()+TRIMOTORLEFT);
	motor[RIGHT].write(Power-control[ROLL].getOutput()+control[YAW].getOutput()+TRIMOTORRIGHT);
}

float Quadcopter::processHeading()
{
	//Based on AeroQuad.com/FlightControl.pde
	
	if (headingHoldConfig == ON) {
		
		heading = degrees(gyro.getHeading());
		
		// Always center relative heading around absolute heading chosen during yaw command
		// This assumes that an incorrect yaw can't be forced on the AeroQuad >180 or <-180 degrees
		// This is done so that AeroQuad does not accidentally hit transition between 0 and 360 or -180 and 180
		// AKA - THERE IS A BUG HERE - if relative heading is greater than 180 degrees, the PID will swing from negative to positive
		// Doubt that will happen as it would have to be uncommanded.
		relativeHeading = heading - setHeading;
		if (heading <= (setHeading - 180)) relativeHeading += 360;
		if (heading >= (setHeading + 180)) relativeHeading -= 360;
		
		// Apply heading hold only when throttle high enough to start flight
		if (Power > 1000 +100 ) { 
			if ((TargetPositionUser[YAW] > (0.125)) || (TargetPositionUser[YAW] < (-0.125))) {
				// If commanding yaw, turn off heading hold and store latest heading
				setHeading = heading;
				headingHold = 0;
				controlHeading.zeroIntegralError();
				headingHoldState = OFF;
				headingTime = currentTime;
			}
			else {
				if (relativeHeading < .25 && relativeHeading > -.25) {
					headingHold = 0;
					controlHeading.zeroIntegralError();
				}					else if (headingHoldState == OFF) { // quick fix to soften heading hold on new heading
					if ((currentTime - headingTime) > 500000) {
						headingHoldState = ON;
						headingTime = currentTime;
						setHeading = heading;
						headingHold = 0;
					}
				}
				else {
					// No new yaw input, calculate current heading vs. desired heading heading hold
					// Relative heading is always centered around zero
					controlHeading.setTargetPosition(0);
					controlHeading.updatePID(relativeHeading);
					headingHold = controlHeading.getOutput();
					headingTime = currentTime; // quick fix to soften heading hold, wait 100ms before applying heading hold
				}
			}
		}
		else {
			// minimum throttle not reached, use off settings
			setHeading = heading;
			headingHold = 0;
			controlHeading.zeroIntegralError();
		}
	}
	
	return radians(headingHold);

}

void Quadcopter::initializeMotors()
{
	motor[FRONT].initialize(FRONTMOTORPIN);
	motor[REAR].initialize(REARMOTORPIN);
	motor[RIGHT].initialize(RIGHTMOTORPIN);
	motor[LEFT].initialize(LEFTMOTORPIN);
	
	delay(TIEMPOARMADOESC);
	disarmed();
}

void Quadcopter::initializeControl()
{
	pitchAttitudeStableMode.setTargetPosition(0);
	rollAttitudeStableMode.setTargetPosition(0);
	control[PITCH].setTargetPosition(0);
	control[ROLL].setTargetPosition(0);
	
	control[YAW].setTargetPosition(0);
	
	// Integral Limit for attitude mode
	// This overrides default set in readEEPROM()
	// Set for 1/2 max attitude command (+/-0.75 radians)
	// Rate integral not used for now
	rollAttitudeStableMode.setWindUpGuard(0.375);
	pitchAttitudeStableMode.setWindUpGuard(0.375);
	
	zeroIntegralError();
	
}

void Quadcopter::initializeComunication()
{	
	//initialize external communication
	comunication.initialize();
}

void Quadcopter::zeroIntegralError()
{
	control[ROLL].zeroIntegralError();
	control[PITCH].zeroIntegralError();
	rollAttitudeStableMode.zeroIntegralError();
	pitchAttitudeStableMode.zeroIntegralError();
	
	control[YAW].zeroIntegralError();
	controlHeading.zeroIntegralError();	
}

void Quadcopter::armed()
{
	motor[FRONT].armed=1;
	motor[REAR].armed=1;
	motor[LEFT].armed=1;
	motor[RIGHT].armed=1;
	
	digitalWrite(LED3PIN,HIGH);
}

void Quadcopter::disarmed()
{
	motor[FRONT].armed=0;
	motor[REAR].armed=0;
	motor[LEFT].armed=0;
	motor[RIGHT].armed=0;
	
	digitalWrite(LED3PIN,LOW);
}

void Quadcopter::initializeReceiver()
{
	receiver.initialize();
}

void Quadcopter::readReceiver()
{
	receiver.read();
	
	//Se hace para hacer leer correctamente los valores del puerto serie cuando el mando esta apagado.
	if((receiver.getData(AUX)>1900 || receiver.getData(THROTTLE)>1060 ||
		receiver.getData(THROTTLE)<1036)&&receiver.getData(THROTTLE)!=0)
	{
	
		if(receiver.getData(AUX)>1900)
			armed();
		else
			disarmed();
		
		// Get center value of roll/pitch/yaw channels when enough throttle to lift off
		if (receiver.getRaw(THROTTLE) < 1300) {
			receiver.setTransmitterTrim(ROLL, receiver.getRaw(ROLL));
			receiver.setTransmitterTrim(PITCH, receiver.getRaw(PITCH));
			receiver.setTransmitterTrim(YAW, receiver.getRaw(YAW));
			//receiver.setZero(ROLL, receiver.getRaw(ROLL));
			//receiver.setZero(PITCH, receiver.getRaw(PITCH));
			//receiver.setZero(YAW, receiver.getRaw(YAW));
		}
		
		//Power
		Power = map(receiver.getData(THROTTLE),1000,2000,1000,MAXRECEIVERTHROTTLE);
	
		TargetPositionUser[PITCH]=(receiver.getData(PITCH) - receiver.getZero(PITCH)) * ATTITUDE_SCALING;
		TargetPositionUser[ROLL]=(receiver.getData(ROLL) - receiver.getZero(ROLL)) * ATTITUDE_SCALING;

		
		//Yaw
		TargetPositionUser[YAW]=receiver.getSIData(YAW);		
	}
	
	//Por ejemplo Control de luces
}

void Quadcopter::readComunication()
{
	// Check for serial message
	if (comunication.available()) 
	{
		digitalWrite(LEDPIN, LOW);
		queryType = comunication.read();
		switch (queryType) {
				
			case 'a': //armed
				armed();
				
				break;
				
			case 'd'://disarmed
				disarmed();
				
				break;

			case 'b': // calibrate gyro
				 gyro.calibrate();
				
				 break;
				 
			case 'c': // calibrate accel
				 accel.calibrate();

				 break;
			
			case 'm': // read Target Position User & Power //By MarotoCopter
				TargetPositionUser[PITCH]=comunication.readFloat();
				TargetPositionUser[ROLL]=comunication.readFloat();
				TargetPositionUser[YAW]=comunication.readFloat();
				//Power
				Power=comunication.readFloat();
					
				break;
				
		}
		digitalWrite(LEDPIN, HIGH);
	}
	
}

void Quadcopter::writeComunication()
{
	switch (queryType) {

	case 'B': // Send roll and pitch gyro PID values
		comunication.write(control[ROLL].getP());
		comunication.write(control[ROLL].getI());
		comunication.write(control[ROLL].getD());
		comunication.write(control[PITCH].getP());
		comunication.write(control[PITCH].getI());
		comunication.writeEnd(control[PITCH].getD());
		queryType = 'X';
	break;
		
	case 'X': // Stop sending messages
	break;

			
	case '&': // Special send Marotocopter_configurator
	//default:	// Special send Marotocopter_configurator
			
				comunication.write(degrees(rollAttitudeStableMode.getTargetPosition()));
				comunication.write(degrees(pitchAttitudeStableMode.getTargetPosition()));
				comunication.write(degrees(rollAttitudeStableMode.getCurrentPosition()));
				comunication.write(degrees(pitchAttitudeStableMode.getCurrentPosition()));
				comunication.write(control[ROLL].getOutput());
				comunication.write(control[PITCH].getOutput());
				comunication.write(control[ROLL].getP());
				comunication.write(control[PITCH].getP());
				comunication.write(control[ROLL].getI());
				comunication.write(control[PITCH].getI());
				comunication.write(control[ROLL].getD());
				comunication.write(control[PITCH].getD());
				comunication.write(accel.getSmoothFactor());
				comunication.write(gyro.getSmoothFactor());
				comunication.writeEnd(Power);
			
			queryType = 'X';
			
	break;
}
}


